@@ -6,12 +6,14 @@
 //! such as `rpm-ostree` can directly reuse it.
 
 use anyhow::Result;
-use ostree::gio;
+use ostree::{gio, glib};
 use std::collections::BTreeMap;
 use std::convert::TryFrom;
 use std::ffi::OsString;
+use std::path::PathBuf;
 use structopt::StructOpt;
 
+use crate::container as ostree_container;
 use crate::container::store::{LayeredImageImporter, PrepareResult};
 use crate::container::{Config, ImageReference, OstreeImageReference, UnencapsulateOptions};
 
@@ -23,18 +25,6 @@ fn parse_base_imgref(s: &str) -> Result<ImageReference> {
     ImageReference::try_from(s)
 }
 
-#[derive(Debug, StructOpt)]
-struct BuildOpts {
-    #[structopt(long)]
-    repo: String,
-
-    #[structopt(long = "ref")]
-    ostree_ref: String,
-
-    #[structopt(long)]
-    oci_dir: String,
-}
-
 /// Options for importing a tar archive.
 #[derive(Debug, StructOpt)]
 struct ImportOpts {
@@ -53,6 +43,10 @@ struct ExportOpts {
     #[structopt(long)]
     repo: String,
 
+    /// The format version.  Must be 0 or 1.
+    #[structopt(long)]
+    format_version: u32,
+
     /// The ostree ref or commit to export
     rev: String,
 }
@@ -124,6 +118,23 @@ enum ContainerOpts {
     Image(ContainerImageOpts),
 }
 
+/// Options for container image fetching.
+#[derive(Debug, StructOpt)]
+struct ContainerProxyOpts {
+    #[structopt(long)]
+    /// Path to Docker-formatted authentication file.
+    authfile: Option<PathBuf>,
+
+    #[structopt(long)]
+    /// Directory with certificates (*.crt, *.cert, *.key) used to connect to registry
+    /// Equivalent to `skopeo --cert-dir`
+    cert_dir: Option<PathBuf>,
+
+    #[structopt(long)]
+    /// Skip TLS verification.
+    insecure_skip_tls_verification: bool,
+}
+
 /// Options for import/export to tar archives.
 #[derive(Debug, StructOpt)]
 enum ContainerImageOpts {
@@ -143,6 +154,9 @@ enum ContainerImageOpts {
         /// Image reference, e.g. ostree-remote-image:someremote:registry:quay.io/exampleos/exampleos:latest
         #[structopt(parse(try_from_str = parse_imgref))]
         imgref: OstreeImageReference,
+
+        #[structopt(flatten)]
+        proxyopts: ContainerProxyOpts,
     },
 
     /// Copy a pulled container image from one repo to another.
@@ -175,6 +189,9 @@ enum ContainerImageOpts {
         #[structopt(parse(try_from_str = parse_imgref))]
         imgref: OstreeImageReference,
 
+        #[structopt(flatten)]
+        proxyopts: ContainerProxyOpts,
+
         /// Target image reference, e.g. ostree-remote-image:someremote:registry:quay.io/exampleos/exampleos:latest
         ///
         /// If specified, `--imgref` will be used as a source, but this reference will be emitted into the origin
@@ -206,10 +223,18 @@ struct ImaSignOpts {
     key: String,
 }
 
+/// Options for internal testing
+#[derive(Debug, StructOpt)]
+enum TestingOpts {
+    // Detect the current environment
+    DetectEnv,
+}
+
 /// Toplevel options for extended ostree functionality.
 #[derive(Debug, StructOpt)]
 #[structopt(name = "ostree-ext")]
 #[structopt(rename_all = "kebab-case")]
+#[allow(clippy::large_enum_variant)]
 enum Opt {
     /// Import and export to tar
     Tar(TarOpts),
@@ -217,6 +242,21 @@ enum Opt {
     Container(ContainerOpts),
     /// IMA signatures
     ImaSign(ImaSignOpts),
+    #[structopt(setting(structopt::clap::AppSettings::Hidden))]
+    #[cfg(feature = "internal-testing-api")]
+    InternalOnlyForTesting(TestingOpts),
+}
+
+#[allow(clippy::from_over_into)]
+impl Into<ostree_container::store::ImageProxyConfig> for ContainerProxyOpts {
+    fn into(self) -> ostree_container::store::ImageProxyConfig {
+        ostree_container::store::ImageProxyConfig {
+            authfile: self.authfile,
+            certificate_directory: self.cert_dir,
+            insecure_skip_tls_verification: Some(self.insecure_skip_tls_verification),
+            ..Default::default()
+        }
+    }
 }
 
 /// Import a tar archive containing an ostree commit.
@@ -236,7 +276,12 @@ async fn tar_import(opts: &ImportOpts) -> Result<()> {
 /// Export a tar archive containing an ostree commit.
 fn tar_export(opts: &ExportOpts) -> Result<()> {
     let repo = &ostree::Repo::open_at(libc::AT_FDCWD, opts.repo.as_str(), gio::NONE_CANCELLABLE)?;
-    crate::tar::export_commit(repo, opts.rev.as_str(), std::io::stdout())?;
+    #[allow(clippy::needless_update)]
+    let subopts = crate::tar::ExportOptions {
+        format_version: opts.format_version,
+        ..Default::default()
+    };
+    crate::tar::export_commit(repo, opts.rev.as_str(), std::io::stdout(), Some(subopts))?;
     Ok(())
 }
 
@@ -264,7 +309,7 @@ async fn container_import(
     let opts = UnencapsulateOptions {
         progress: Some(tx_progress),
     };
-    let import = crate::container::unencapsulate(repo, &imgref, Some(opts));
+    let import = crate::container::unencapsulate(repo, imgref, Some(opts));
     tokio::pin!(import);
     tokio::pin!(rx_progress);
     let import = loop {
@@ -316,25 +361,30 @@ async fn container_export(
         labels: Some(labels),
         cmd,
     };
-    let pushed = crate::container::encapsulate(repo, rev, &config, &imgref).await?;
+    let opts = Some(Default::default());
+    let pushed = crate::container::encapsulate(repo, rev, &config, opts, imgref).await?;
     println!("{}", pushed);
     Ok(())
 }
 
 /// Load metadata for a container image with an encapsulated ostree commit.
 async fn container_info(imgref: &OstreeImageReference) -> Result<()> {
-    let (_, digest) = crate::container::fetch_manifest(&imgref).await?;
+    let (_, digest) = crate::container::fetch_manifest(imgref).await?;
     println!("{} digest: {}", imgref, digest);
     Ok(())
 }
 
 /// Write a layered container image into an OSTree commit.
-async fn container_store(repo: &str, imgref: &OstreeImageReference) -> Result<()> {
+async fn container_store(
+    repo: &str,
+    imgref: &OstreeImageReference,
+    proxyopts: ContainerProxyOpts,
+) -> Result<()> {
     let repo = &ostree::Repo::open_at(libc::AT_FDCWD, repo, gio::NONE_CANCELLABLE)?;
-    let mut imp = LayeredImageImporter::new(repo, &imgref).await?;
+    let mut imp = LayeredImageImporter::new(repo, imgref, proxyopts.into()).await?;
     let prep = match imp.prepare().await? {
         PrepareResult::AlreadyPresent(c) => {
-            println!("No changes in {} => {}", imgref, c);
+            println!("No changes in {} => {}", imgref, c.merge_commit);
             return Ok(());
         }
         PrepareResult::Ready(r) => r,
@@ -358,18 +408,20 @@ async fn container_store(repo: &str, imgref: &OstreeImageReference) -> Result<()
         }
     }
     let import = imp.import(prep).await?;
-    if !import.layer_filtered_content.is_empty() {
-        for (layerid, filtered) in import.layer_filtered_content {
+    let commit = &repo.load_commit(&import.merge_commit)?.0;
+    let commit_meta = &glib::VariantDict::new(Some(&commit.child_value(0)));
+    let filtered = commit_meta.lookup::<ostree_container::store::MetaFilteredData>(
+        ostree_container::store::META_FILTERED,
+    )?;
+    if let Some(filtered) = filtered {
+        for (layerid, filtered) in filtered {
             eprintln!("Unsupported paths filtered from {}:", layerid);
             for (prefix, count) in filtered {
                 eprintln!("  {}: {}", prefix, count);
             }
         }
     }
-    println!(
-        "Wrote: {} => {} => {}",
-        imgref, import.ostree_ref, import.commit
-    );
+    println!("Wrote: {} => {}", imgref, import.merge_commit);
     Ok(())
 }
 
@@ -392,6 +444,17 @@ fn ima_sign(cmdopts: &ImaSignOpts) -> Result<()> {
     Ok(())
 }
 
+#[cfg(feature = "internal-testing-api")]
+fn testing(opts: &TestingOpts) -> Result<()> {
+    match opts {
+        TestingOpts::DetectEnv => {
+            let s = crate::integrationtest::detectenv();
+            println!("{}", s);
+            Ok(())
+        }
+    }
+}
+
 /// Parse the provided arguments and execute.
 /// Calls [`structopt::clap::Error::exit`] on failure, printing the error message and aborting the program.
 pub async fn run_from_iter<I>(args: I) -> Result<()>
@@ -421,10 +484,8 @@ where
                 let labels: Result<BTreeMap<_, _>> = labels
                     .into_iter()
                     .map(|l| {
-                        let mut parts = l.splitn(2, '=');
-                        let k = parts.next().unwrap();
-                        let v = parts
-                            .next()
+                        let (k, v) = l
+                            .split_once('=')
                             .ok_or_else(|| anyhow::anyhow!("Missing '=' in label {}", l))?;
                         Ok((k.to_string(), v.to_string()))
                     })
@@ -440,7 +501,11 @@ where
                     }
                     Ok(())
                 }
-                ContainerImageOpts::Pull { repo, imgref } => container_store(&repo, &imgref).await,
+                ContainerImageOpts::Pull {
+                    repo,
+                    imgref,
+                    proxyopts,
+                } => container_store(&repo, &imgref, proxyopts).await,
                 ContainerImageOpts::Copy {
                     src_repo,
                     dest_repo,
@@ -458,6 +523,7 @@ where
                     imgref,
                     target_imgref,
                     karg,
+                    proxyopts,
                 } => {
                     let sysroot = &ostree::Sysroot::new(Some(&gio::File::for_path(&sysroot)));
                     sysroot.load(gio::NONE_CANCELLABLE)?;
@@ -469,6 +535,7 @@ where
                     let options = crate::container::deploy::DeployOpts {
                         kargs: kargs.as_deref(),
                         target_imgref: target_imgref.as_ref(),
+                        proxy_cfg: Some(proxyopts.into()),
                     };
                     crate::container::deploy::deploy(sysroot, &stateroot, &imgref, Some(options))
                         .await
@@ -476,5 +543,7 @@ where
             },
         },
         Opt::ImaSign(ref opts) => ima_sign(opts),
+        #[cfg(feature = "internal-testing-api")]
+        Opt::InternalOnlyForTesting(ref opts) => testing(opts),
     }
 }
