@@ -10,10 +10,17 @@ use crate::refescape;
 use anyhow::{anyhow, Context};
 use containers_image_proxy::{ImageProxy, OpenedImage};
 use fn_error_context::context;
-use oci_spec::image as oci_image;
+use oci_spec::image::{self as oci_image, ImageManifest};
 use ostree::prelude::{Cast, ToVariant};
 use ostree::{gio, glib};
-use std::collections::{BTreeMap, HashMap};
+use std::collections::HashMap;
+use std::iter::FromIterator;
+
+/// Configuration for the proxy.
+///
+/// We re-export this rather than inventing our own wrapper
+/// in the interest of avoiding duplication.
+pub use containers_image_proxy::ImageProxyConfig;
 
 /// The ostree ref prefix for blobs.
 const LAYER_PREFIX: &str = "ostree/container/blob";
@@ -24,6 +31,12 @@ const IMAGE_PREFIX: &str = "ostree/container/image";
 const META_MANIFEST_DIGEST: &str = "ostree.manifest-digest";
 /// The key injected into the merge commit with the manifest serialized as JSON.
 const META_MANIFEST: &str = "ostree.manifest";
+/// The key injected into the merge commit with the image configuration serialized as JSON.
+const META_CONFIG: &str = "ostree.container.image-config";
+/// Value of type `a{sa{su}}` containing number of filtered out files
+pub const META_FILTERED: &str = "ostree.tar-filtered";
+/// The type used to store content filtering information with `META_FILTERED`.
+pub type MetaFilteredData = HashMap<String, HashMap<String, u32>>;
 
 /// Convert e.g. sha256:12345... into `/ostree/container/blob/sha256_2B12345...`.
 fn ref_for_blob_digest(d: &str) -> Result<String> {
@@ -40,19 +53,50 @@ fn ref_for_image(l: &ImageReference) -> Result<String> {
     refescape::prefix_escape_for_ref(IMAGE_PREFIX, &l.to_string())
 }
 
+/// State of an already pulled layered image.
+#[derive(Debug, PartialEq, Eq)]
+pub struct LayeredImageState {
+    /// The base ostree commit
+    pub base_commit: String,
+    /// The merge commit unions all layers
+    pub merge_commit: String,
+    /// Whether or not the image has multiple layers.
+    pub is_layered: bool,
+    /// The digest of the original manifest
+    pub manifest_digest: String,
+}
+
+impl LayeredImageState {
+    /// Return the default ostree commit digest for this image.
+    ///
+    /// If this is a non-layered image, the merge commit will be
+    /// ignored, and the base commit returned.
+    ///
+    /// Otherwise, this returns the merge commit.
+    pub fn get_commit(&self) -> &str {
+        if self.is_layered {
+            self.merge_commit.as_str()
+        } else {
+            self.base_commit.as_str()
+        }
+    }
+}
+
 /// Context for importing a container image.
+#[derive(Debug)]
 pub struct LayeredImageImporter {
     repo: ostree::Repo,
     proxy: ImageProxy,
     imgref: OstreeImageReference,
+    target_imgref: Option<OstreeImageReference>,
     proxy_img: OpenedImage,
-    ostree_ref: String,
 }
 
 /// Result of invoking [`LayeredImageImporter::prepare`].
+#[derive(Debug)]
 pub enum PrepareResult {
     /// The image reference is already present; the contained string is the OSTree commit.
-    AlreadyPresent(String),
+    AlreadyPresent(LayeredImageState),
     /// The image needs to be downloaded
     Ready(Box<PreparedImport>),
 }
@@ -86,6 +130,8 @@ pub struct PreparedImport {
     pub manifest_digest: String,
     /// The deserialized manifest.
     pub manifest: oci_image::ImageManifest,
+    /// The deserialized configuration.
+    pub config: Option<oci_image::ImageConfiguration>,
     /// The previously stored manifest digest.
     pub previous_manifest_digest: Option<String>,
     /// The previously stored image ID.
@@ -96,18 +142,6 @@ pub struct PreparedImport {
     pub layers: Vec<ManifestLayerState>,
 }
 
-/// A successful import of a container image.
-#[derive(Debug, PartialEq, Eq)]
-pub struct CompletedImport {
-    /// The ostree ref used for the container image.
-    pub ostree_ref: String,
-    /// The current commit.
-    pub commit: String,
-    /// A mapping from layer blob IDs to a count of content filtered out
-    /// by toplevel path.
-    pub layer_filtered_content: BTreeMap<String, BTreeMap<String, u32>>,
-}
-
 // Given a manifest, compute its ostree ref name and cached ostree commit
 fn query_layer(repo: &ostree::Repo, layer: oci_image::Descriptor) -> Result<ManifestLayerState> {
     let ostree_ref = ref_for_layer(&layer)?;
@@ -145,20 +179,28 @@ pub fn manifest_digest_from_commit(commit: &glib::Variant) -> Result<String> {
 
 impl LayeredImageImporter {
     /// Create a new importer.
-    pub async fn new(repo: &ostree::Repo, imgref: &OstreeImageReference) -> Result<Self> {
-        let proxy = ImageProxy::new().await?;
+    pub async fn new(
+        repo: &ostree::Repo,
+        imgref: &OstreeImageReference,
+        config: ImageProxyConfig,
+    ) -> Result<Self> {
+        let proxy = ImageProxy::new_with_config(config).await?;
         let proxy_img = proxy.open_image(&imgref.imgref.to_string()).await?;
         let repo = repo.clone();
-        let ostree_ref = ref_for_image(&imgref.imgref)?;
         Ok(LayeredImageImporter {
             repo,
             proxy,
             proxy_img,
-            ostree_ref,
+            target_imgref: None,
             imgref: imgref.clone(),
         })
     }
 
+    /// Write cached data as if the image came from this source.
+    pub fn set_target(&mut self, target: &OstreeImageReference) {
+        self.target_imgref = Some(target.clone())
+    }
+
     /// Determine if there is a new manifest, and if so return its digest.
     #[context("Fetching manifest")]
     pub async fn prepare(&mut self) -> Result<PrepareResult> {
@@ -175,30 +217,42 @@ impl LayeredImageImporter {
         }
 
         let (manifest_digest, manifest_bytes) = self.proxy.fetch_manifest(&self.proxy_img).await?;
-        let manifest: oci_image::ImageManifest = serde_json::from_slice(&manifest_bytes)?;
+        let manifest: oci_image::ImageManifest =
+            serde_json::from_slice(&manifest_bytes).context("Parsing image manifest")?;
         let new_imageid = manifest.config().digest().as_str();
 
         // Query for previous stored state
-        let (previous_manifest_digest, previous_imageid) = if let Some(merge_commit) =
-            self.repo.resolve_rev(&self.ostree_ref, true)?
-        {
-            let merge_commit_obj = &self.repo.load_commit(merge_commit.as_str())?.0;
-            let commit_meta = &merge_commit_obj.child_value(0);
-            let commit_meta = &ostree::glib::VariantDict::new(Some(commit_meta));
-            let (previous_manifest, previous_digest) = manifest_data_from_commitmeta(commit_meta)?;
-            // If the manifest digests match, we're done.
-            if previous_digest == manifest_digest {
-                return Ok(PrepareResult::AlreadyPresent(merge_commit.to_string()));
-            }
-            // Failing that, if they have the same imageID, we're also done.
-            let previous_imageid = previous_manifest.config().digest().as_str();
-            if previous_imageid == new_imageid {
-                return Ok(PrepareResult::AlreadyPresent(merge_commit.to_string()));
-            }
-            (Some(previous_digest), Some(previous_imageid.to_string()))
-        } else {
-            (None, None)
+
+        let (previous_manifest_digest, previous_imageid) =
+            if let Some((previous_manifest, previous_state)) =
+                query_image_impl(&self.repo, &self.imgref)?
+            {
+                // If the manifest digests match, we're done.
+                if previous_state.manifest_digest == manifest_digest {
+                    return Ok(PrepareResult::AlreadyPresent(previous_state));
+                }
+                // Failing that, if they have the same imageID, we're also done.
+                let previous_imageid = previous_manifest.config().digest().as_str();
+                if previous_imageid == new_imageid {
+                    return Ok(PrepareResult::AlreadyPresent(previous_state));
+                }
+                (
+                    Some(previous_state.manifest_digest),
+                    Some(previous_imageid.to_string()),
+                )
+            } else {
+                (None, None)
+            };
+
+        #[cfg(feature = "proxy_v0_2_3")]
+        let config = {
+            let config_bytes = self.proxy.fetch_config(&self.proxy_img).await?;
+            let config: oci_image::ImageConfiguration =
+                serde_json::from_slice(&config_bytes).context("Parsing image configuration")?;
+            Some(config)
         };
+        #[cfg(not(feature = "proxy_v0_2_3"))]
+        let config = None;
 
         let mut layers = manifest.layers().iter().cloned();
         // We require a base layer.
@@ -213,6 +267,7 @@ impl LayeredImageImporter {
         let imp = PreparedImport {
             manifest,
             manifest_digest,
+            config,
             previous_manifest_digest,
             previous_imageid,
             base_layer,
@@ -222,44 +277,48 @@ impl LayeredImageImporter {
     }
 
     /// Import a layered container image
-    pub async fn import(self, import: Box<PreparedImport>) -> Result<CompletedImport> {
-        let proxy = self.proxy;
+    pub async fn import(self, import: Box<PreparedImport>) -> Result<LayeredImageState> {
+        let mut proxy = self.proxy;
+        let target_imgref = self.target_imgref.as_ref().unwrap_or(&self.imgref);
+
         // First download the base image (if necessary) - we need the SELinux policy
         // there to label all following layers.
         let base_layer = import.base_layer;
         let base_commit = if let Some(c) = base_layer.commit {
             c
         } else {
-            let base_layer_ref = &base_layer.layer;
-            let (blob, driver) = super::unencapsulate::fetch_layer_decompress(
-                &proxy,
+            let base_commit = super::unencapsulate_from_manifest_impl(
+                &self.repo,
+                &mut proxy,
+                target_imgref,
                 &self.proxy_img,
-                &base_layer.layer,
+                &import.manifest,
+                None,
+                true,
             )
             .await?;
-            let importer = crate::tar::import_tar(&self.repo, blob, None);
-            let (commit, driver) = tokio::join!(importer, driver);
-            driver?;
-            let commit =
-                commit.with_context(|| format!("Parsing blob {}", base_layer_ref.digest()))?;
-            // TODO support ref writing in tar import
+            // Write the ostree ref for that single layer; TODO
+            // handle this as part of the overall transaction.
             self.repo.set_ref_immediate(
                 None,
                 base_layer.ostree_ref.as_str(),
-                Some(commit.as_str()),
+                Some(base_commit.as_str()),
                 gio::NONE_CANCELLABLE,
             )?;
-            commit
+            base_commit
         };
 
+        let ostree_ref = ref_for_image(&target_imgref.imgref)?;
+
         let mut layer_commits = Vec::new();
-        let mut layer_filtered_content = BTreeMap::new();
+        let mut layer_filtered_content: MetaFilteredData = HashMap::new();
         for layer in import.layers {
             if let Some(c) = layer.commit {
+                tracing::debug!("Reusing fetched commit {}", c);
                 layer_commits.push(c.to_string());
             } else {
                 let (blob, driver) = super::unencapsulate::fetch_layer_decompress(
-                    &proxy,
+                    &mut proxy,
                     &self.proxy_img,
                     &layer.layer,
                 )
@@ -272,34 +331,40 @@ impl LayeredImageImporter {
                 };
                 let w =
                     crate::tar::write_tar(&self.repo, blob, layer.ostree_ref.as_str(), Some(opts));
-                let (r, driver) = tokio::join!(w, driver);
-                let r = r.with_context(|| format!("Parsing layer blob {}", layer.digest()))?;
-                driver?;
+                let r = super::unencapsulate::join_fetch(w, driver)
+                    .await
+                    .with_context(|| format!("Parsing layer blob {}", layer.digest()))?;
                 layer_commits.push(r.commit);
                 if !r.filtered.is_empty() {
-                    layer_filtered_content.insert(layer.digest().to_string(), r.filtered);
+                    let filtered = HashMap::from_iter(r.filtered.into_iter());
+                    layer_filtered_content.insert(layer.digest().to_string(), filtered);
                 }
             }
         }
 
         // We're done with the proxy, make sure it didn't have any errors.
         proxy.finalize().await?;
+        tracing::debug!("finalized proxy");
 
         let serialized_manifest = serde_json::to_string(&import.manifest)?;
+        let serialized_config = serde_json::to_string(&import.config)?;
         let mut metadata = HashMap::new();
         metadata.insert(META_MANIFEST_DIGEST, import.manifest_digest.to_variant());
         metadata.insert(META_MANIFEST, serialized_manifest.to_variant());
+        metadata.insert(META_CONFIG, serialized_config.to_variant());
         metadata.insert(
             "ostree.importer.version",
             env!("CARGO_PKG_VERSION").to_variant(),
         );
+        let filtered = layer_filtered_content.to_variant();
+        metadata.insert(META_FILTERED, filtered);
         let metadata = metadata.to_variant();
 
         // Destructure to transfer ownership to thread
         let repo = self.repo;
-        let target_ref = self.ostree_ref;
-        let (ostree_ref, commit) = crate::tokio_util::spawn_blocking_cancellable(
-            move |cancellable| -> Result<(String, String)> {
+        let imgref = self.target_imgref.unwrap_or(self.imgref);
+        let state = crate::tokio_util::spawn_blocking_cancellable(
+            move |cancellable| -> Result<LayeredImageState> {
                 let cancellable = Some(cancellable);
                 let repo = &repo;
                 let txn = repo.auto_transaction(cancellable)?;
@@ -328,17 +393,16 @@ impl LayeredImageImporter {
                     &merged_root,
                     cancellable,
                 )?;
-                repo.transaction_set_ref(None, &target_ref, Some(merged_commit.as_str()));
+                repo.transaction_set_ref(None, &ostree_ref, Some(merged_commit.as_str()));
                 txn.commit(cancellable)?;
-                Ok((target_ref, merged_commit.to_string()))
+                // Here we re-query state just to run through the same code path,
+                // though it'd be cheaper to synthesize it from the data we already have.
+                let state = query_image(repo, &imgref)?.unwrap();
+                Ok(state)
             },
         )
         .await??;
-        Ok(CompletedImport {
-            ostree_ref,
-            commit,
-            layer_filtered_content,
-        })
+        Ok(state)
     }
 }
 
@@ -355,6 +419,47 @@ pub fn list_images(repo: &ostree::Repo) -> Result<Vec<String>> {
         .collect()
 }
 
+fn query_image_impl(
+    repo: &ostree::Repo,
+    imgref: &OstreeImageReference,
+) -> Result<Option<(ImageManifest, LayeredImageState)>> {
+    let ostree_ref = &ref_for_image(&imgref.imgref)?;
+    let merge_rev = repo.resolve_rev(ostree_ref, true)?;
+    let (merge_commit, merge_commit_obj) = if let Some(r) = merge_rev {
+        (r.to_string(), repo.load_commit(r.as_str())?.0)
+    } else {
+        return Ok(None);
+    };
+    let commit_meta = &merge_commit_obj.child_value(0);
+    let commit_meta = &ostree::glib::VariantDict::new(Some(commit_meta));
+    let (manifest, manifest_digest) = manifest_data_from_commitmeta(commit_meta)?;
+    let mut layers = manifest.layers().iter().cloned();
+    // We require a base layer.
+    let base_layer = layers.next().ok_or_else(|| anyhow!("No layers found"))?;
+    let base_layer = query_layer(repo, base_layer)?;
+    let base_commit = base_layer
+        .commit
+        .ok_or_else(|| anyhow!("Missing base image ref"))?;
+    // If there are more layers after the base, then we're layered.
+    let is_layered = layers.count() > 0;
+    let state = LayeredImageState {
+        base_commit,
+        merge_commit,
+        is_layered,
+        manifest_digest,
+    };
+    tracing::debug!(state = ?state);
+    Ok(Some((manifest, state)))
+}
+
+/// Query metadata for a pulled image.
+pub fn query_image(
+    repo: &ostree::Repo,
+    imgref: &OstreeImageReference,
+) -> Result<Option<LayeredImageState>> {
+    Ok(query_image_impl(repo, imgref)?.map(|v| v.1))
+}
+
 /// Copy a downloaded image from one repository to another.
 pub async fn copy(
     src_repo: &ostree::Repo,
