@@ -4,8 +4,10 @@
 //! and the Rust bindings to it, adding new functionality
 //! written in Rust.  
 
+// See https://doc.rust-lang.org/rustc/lints/listing/allowed-by-default.html
 #![deny(missing_docs)]
-// Good defaults
+#![deny(missing_debug_implementations)]
+#![deny(unreachable_pub)]
 #![forbid(unused_must_use)]
 #![deny(unsafe_code)]
 #![cfg_attr(feature = "dox", feature(doc_cfg))]
@@ -22,9 +24,9 @@ pub use ostree::gio::glib;
 /// to a string to output to a terminal or logs.
 type Result<T> = anyhow::Result<T>;
 
-mod async_util;
 pub mod cli;
 pub mod container;
+pub mod container_utils;
 pub mod diff;
 pub mod ima;
 pub mod keyfileext;
@@ -39,3 +41,6 @@ pub mod prelude {
     #[doc(hidden)]
     pub use ostree::prelude::*;
 }
+
+#[cfg(feature = "internal-testing-api")]
+mod integrationtest;
