@@ -2,6 +2,7 @@ use anyhow::{Context, Result};
 use camino::{Utf8Path, Utf8PathBuf};
 use fn_error_context::context;
 use indoc::indoc;
+use once_cell::sync::Lazy;
 use ostree_ext::container::store::PrepareResult;
 use ostree_ext::container::{
     Config, ImageReference, OstreeImageReference, SignatureSource, Transport,
@@ -35,14 +36,10 @@ fn assert_err_contains<T>(r: Result<T>, s: impl AsRef<str>) {
     }
 }
 
-lazy_static::lazy_static! {
-    static ref TEST_REGISTRY: String = {
-        match std::env::var_os("TEST_REGISTRY") {
-            Some(t) => t.to_str().unwrap().to_owned(),
-            None => TEST_REGISTRY_DEFAULT.to_string()
-        }
-    };
-}
+static TEST_REGISTRY: Lazy<String> = Lazy::new(|| match std::env::var_os("TEST_REGISTRY") {
+    Some(t) => t.to_str().unwrap().to_owned(),
+    None => TEST_REGISTRY_DEFAULT.to_string(),
+});
 
 #[context("Generating test repo")]
 fn generate_test_repo(dir: &Utf8Path) -> Result<Utf8PathBuf> {
@@ -103,7 +100,11 @@ fn initial_export(fixture: &Fixture) -> Result<Utf8PathBuf> {
     );
     let destpath = fixture.path.join("exampleos-export.tar");
     let mut outf = std::io::BufWriter::new(std::fs::File::create(&destpath)?);
-    ostree_ext::tar::export_commit(&fixture.srcrepo, rev.as_str(), &mut outf)?;
+    let options = ostree_ext::tar::ExportOptions {
+        format_version: fixture.format_version,
+        ..Default::default()
+    };
+    ostree_ext::tar::export_commit(&fixture.srcrepo, rev.as_str(), &mut outf, Some(options))?;
     outf.flush()?;
     Ok(destpath)
 }
@@ -116,6 +117,8 @@ struct Fixture {
     srcrepo: ostree::Repo,
     destrepo: ostree::Repo,
     destrepo_path: Utf8PathBuf,
+
+    format_version: u32,
 }
 
 impl Fixture {
@@ -142,6 +145,7 @@ impl Fixture {
             srcrepo,
             destrepo,
             destrepo_path,
+            format_version: 0,
         })
     }
 }
@@ -162,14 +166,14 @@ async fn test_tar_export_reproducible() -> Result<()> {
         .read_commit(TESTREF, gio::NONE_CANCELLABLE)?;
     let export1 = {
         let mut h = openssl::hash::Hasher::new(openssl::hash::MessageDigest::sha256())?;
-        ostree_ext::tar::export_commit(&fixture.srcrepo, rev.as_str(), &mut h)?;
+        ostree_ext::tar::export_commit(&fixture.srcrepo, rev.as_str(), &mut h, None)?;
         h.finish()?
     };
     // Artificial delay to flush out mtimes (one second granularity baseline, plus another 100ms for good measure).
     std::thread::sleep(std::time::Duration::from_millis(1100));
     let export2 = {
         let mut h = openssl::hash::Hasher::new(openssl::hash::MessageDigest::sha256())?;
-        ostree_ext::tar::export_commit(&fixture.srcrepo, rev.as_str(), &mut h)?;
+        ostree_ext::tar::export_commit(&fixture.srcrepo, rev.as_str(), &mut h, None)?;
         h.finish()?
     };
     assert_eq!(*export1, *export2);
@@ -226,13 +230,95 @@ async fn test_tar_import_signed() -> Result<()> {
         }),
     )
     .await?;
-    let (commitdata, _) = fixture.destrepo.load_commit(&imported)?;
+    let (commitdata, state) = fixture.destrepo.load_commit(&imported)?;
     assert_eq!(
         EXAMPLEOS_CONTENT_CHECKSUM,
         ostree::commit_get_content_checksum(&commitdata)
             .unwrap()
             .as_str()
     );
+    assert_eq!(state, ostree::RepoCommitState::NORMAL);
+    Ok(())
+}
+
+struct TarExpected {
+    path: &'static str,
+    etype: tar::EntryType,
+    mode: u32,
+}
+
+impl Into<TarExpected> for &(&'static str, tar::EntryType, u32) {
+    fn into(self) -> TarExpected {
+        TarExpected {
+            path: self.0,
+            etype: self.1,
+            mode: self.2,
+        }
+    }
+}
+
+fn validate_tar_expected<T: std::io::Read>(
+    t: tar::Entries<T>,
+    expected: impl IntoIterator<Item = TarExpected>,
+) -> Result<()> {
+    let expected = expected.into_iter();
+    let mut entries = t.map(|e| e.unwrap());
+    // Verify we're injecting directories, fixes the absence of `/tmp` in our
+    // images for example.
+    for exp in expected {
+        let mut found = false;
+        while let Some(entry) = entries.next() {
+            let header = entry.header();
+            let entry_path = entry.path().unwrap();
+            if exp.path == entry_path.as_os_str() {
+                assert_eq!(header.entry_type(), exp.etype);
+                assert_eq!(header.mode().unwrap(), exp.mode);
+                found = true;
+                break;
+            }
+        }
+        if !found {
+            anyhow::bail!("Failed to find entry: {}", exp.path);
+        }
+    }
+    Ok(())
+}
+
+/// Validate basic structure of the tar export.
+/// Right now just checks the first entry is `sysroot` with mode 0755.
+#[test]
+fn test_tar_export_structure() -> Result<()> {
+    let mut fixture = Fixture::new()?;
+    let src_tar = initial_export(&fixture)?;
+    let src_tar = std::io::BufReader::new(std::fs::File::open(&src_tar)?);
+    let mut src_tar = tar::Archive::new(src_tar);
+    let mut entries = src_tar.entries()?;
+    // The first entry should be the root directory.
+    let first = entries.next().unwrap()?;
+    let firstpath = first.path()?;
+    assert_eq!(firstpath.to_str().unwrap(), "./");
+    assert_eq!(first.header().mode()?, libc::S_IFDIR | 0o755);
+    let next = entries.next().unwrap().unwrap();
+    assert_eq!(next.path().unwrap().as_os_str(), "sysroot");
+
+    // Validate format version 0
+    let expected = [
+        ("sysroot/config", tar::EntryType::Regular, 0o644),
+        ("usr", tar::EntryType::Directory, libc::S_IFDIR | 0o755),
+    ];
+    validate_tar_expected(entries, expected.iter().map(Into::into))?;
+
+    // Validate format version 1
+    fixture.format_version = 1;
+    let src_tar = initial_export(&fixture)?;
+    let src_tar = std::io::BufReader::new(std::fs::File::open(&src_tar)?);
+    let mut src_tar = tar::Archive::new(src_tar);
+    let expected = [
+        ("sysroot/ostree/repo/config", tar::EntryType::Regular, 0o644),
+        ("usr", tar::EntryType::Directory, libc::S_IFDIR | 0o755),
+    ];
+    validate_tar_expected(src_tar.entries()?, expected.iter().map(Into::into))?;
+
     Ok(())
 }
 
@@ -326,10 +412,15 @@ async fn test_container_import_export() -> Result<()> {
         ),
         cmd: Some(vec!["/bin/bash".to_string()]),
     };
-    let digest =
-        ostree_ext::container::encapsulate(&fixture.srcrepo, TESTREF, &config, &srcoci_imgref)
-            .await
-            .context("exporting")?;
+    let digest = ostree_ext::container::encapsulate(
+        &fixture.srcrepo,
+        TESTREF,
+        &config,
+        None,
+        &srcoci_imgref,
+    )
+    .await
+    .context("exporting")?;
     assert!(srcoci_path.exists());
 
     let inspect = skopeo_inspect(&srcoci_imgref.to_string())?;
@@ -426,9 +517,12 @@ async fn test_container_write_derive() -> Result<()> {
     assert!(images.is_empty());
 
     // Pull a derived image - two layers, new base plus one layer.
-    let mut imp =
-        ostree_ext::container::store::LayeredImageImporter::new(&fixture.destrepo, &exampleos_ref)
-            .await?;
+    let mut imp = ostree_ext::container::store::LayeredImageImporter::new(
+        &fixture.destrepo,
+        &exampleos_ref,
+        Default::default(),
+    )
+    .await?;
     let prep = match imp.prepare().await? {
         PrepareResult::AlreadyPresent(_) => panic!("should not be already imported"),
         PrepareResult::Ready(r) => r,
@@ -444,35 +538,56 @@ async fn test_container_write_derive() -> Result<()> {
     assert_eq!(images.len(), 1);
     assert_eq!(images[0], exampleos_ref.imgref.to_string());
 
-    let imported_commit = &fixture.destrepo.load_commit(import.commit.as_str())?.0;
+    let imported_commit = &fixture
+        .destrepo
+        .load_commit(import.merge_commit.as_str())?
+        .0;
     let digest = ostree_ext::container::store::manifest_digest_from_commit(imported_commit)?;
     assert!(digest.starts_with("sha256:"));
     assert_eq!(digest, expected_digest);
 
+    #[cfg(feature = "proxy_v0_2_3")]
+    {
+        let commit_meta = &imported_commit.child_value(0);
+        let proxy = containers_image_proxy::ImageProxy::new().await?;
+        let commit_meta = glib::VariantDict::new(Some(commit_meta));
+        let config = commit_meta
+            .lookup::<String>("ostree.container.image-config")?
+            .unwrap();
+        let config: oci_spec::image::ImageConfiguration = serde_json::from_str(&config)?;
+        assert_eq!(config.os(), &oci_spec::image::Os::Linux);
+    }
+
     // Parse the commit and verify we pulled the derived content.
     bash!(
         "ostree --repo={repo} ls {r} /usr/share/anewfile",
         repo = fixture.destrepo_path.as_str(),
-        r = import.ostree_ref.as_str()
+        r = import.merge_commit.as_str()
     )?;
 
     // Import again, but there should be no changes.
-    let mut imp =
-        ostree_ext::container::store::LayeredImageImporter::new(&fixture.destrepo, &exampleos_ref)
-            .await?;
+    let mut imp = ostree_ext::container::store::LayeredImageImporter::new(
+        &fixture.destrepo,
+        &exampleos_ref,
+        Default::default(),
+    )
+    .await?;
     let already_present = match imp.prepare().await? {
         PrepareResult::AlreadyPresent(c) => c,
         PrepareResult::Ready(_) => {
-            panic!("Should have already imported {}", import.ostree_ref)
+            panic!("Should have already imported {}", &exampleos_ref)
         }
     };
-    assert_eq!(import.commit, already_present);
+    assert_eq!(import.merge_commit, already_present.merge_commit);
 
     // Test upgrades; replace the oci-archive with new content.
     std::fs::write(exampleos_path, EXAMPLEOS_DERIVED_V2_OCI)?;
-    let mut imp =
-        ostree_ext::container::store::LayeredImageImporter::new(&fixture.destrepo, &exampleos_ref)
-            .await?;
+    let mut imp = ostree_ext::container::store::LayeredImageImporter::new(
+        &fixture.destrepo,
+        &exampleos_ref,
+        Default::default(),
+    )
+    .await?;
     let prep = match imp.prepare().await? {
         PrepareResult::AlreadyPresent(_) => panic!("should not be already imported"),
         PrepareResult::Ready(r) => r,
@@ -486,7 +601,7 @@ async fn test_container_write_derive() -> Result<()> {
     }
     let import = imp.import(prep).await?;
     // New commit.
-    assert_ne!(import.commit, already_present);
+    assert_ne!(import.merge_commit, already_present.merge_commit);
     // We should still have exactly one image stored.
     let images = ostree_ext::container::store::list_images(&fixture.destrepo)?;
     assert_eq!(images.len(), 1);
@@ -500,20 +615,23 @@ async fn test_container_write_derive() -> Result<()> {
          fi
         ",
         repo = fixture.destrepo_path.as_str(),
-        r = import.ostree_ref.as_str()
+        r = import.merge_commit.as_str()
     )?;
 
     // And there should be no changes on upgrade again.
-    let mut imp =
-        ostree_ext::container::store::LayeredImageImporter::new(&fixture.destrepo, &exampleos_ref)
-            .await?;
+    let mut imp = ostree_ext::container::store::LayeredImageImporter::new(
+        &fixture.destrepo,
+        &exampleos_ref,
+        Default::default(),
+    )
+    .await?;
     let already_present = match imp.prepare().await? {
         PrepareResult::AlreadyPresent(c) => c,
         PrepareResult::Ready(_) => {
-            panic!("Should have already imported {}", import.ostree_ref)
+            panic!("Should have already imported {}", &exampleos_ref)
         }
     };
-    assert_eq!(import.commit, already_present);
+    assert_eq!(import.merge_commit, already_present.merge_commit);
 
     // Create a new repo, and copy to it
     let destrepo2 = ostree::Repo::create_at(
@@ -557,7 +675,7 @@ async fn test_container_import_export_registry() -> Result<()> {
         ..Default::default()
     };
     let digest =
-        ostree_ext::container::encapsulate(&fixture.srcrepo, TESTREF, &config, &src_imgref)
+        ostree_ext::container::encapsulate(&fixture.srcrepo, TESTREF, &config, None, &src_imgref)
             .await
             .context("exporting to registry")?;
     let mut digested_imgref = src_imgref.clone();
